# Qorus in Docker

There are three images available, one based on Ubuntu Focal Fossa (20.04 LTS), and two baed on Alpine Linux.  Docker images contain a minimal set of packages able to run everything.  All Docker images as of Qorus v5.1+ are multiarch images, supporting amd64 (x86_64) and arm64 (aarch64) processors.

Available images:
- dregistry.qoretechnologies.com/qorus:5.1 (based on Ubuntu Focal Fossa - 20.04 LTS)
- dregistry.qoretechnologies.com/qorus-alpine:5.1 (based on Alpine Linux)
- dregistry.qoretechnologies.com/qorus-alpine-minimal:5.1 (based on Alpine Linux, without the Oracle client libraries)

## Pulling

In order to use the image you have to first pull it:
```
docker pull dregistry.qoretechnologies.com/qorus:5.1
```

## Usage

To run a Qorus container in the background, you can run a command similar to the following:

```
docker run --name <CONTAINER> \
    --link <DB_CONTAINER>:<DB_IMAGE> \
    --restart always
    -p 8011:8011 \
    -e OMQ_DB_NAME=postgres \
    -e OMQ_DB_TYPE=pgsql \
    -e OMQ_DB_HOST=postgres \
    -e OMQ_DB_USER=postgres \
    -e OMQ_DB_PASS=omq \
    -e TZ=Europe/Prague \
    -e QORUS_ADMIN_PASS=changeme \
    -e QORUS_UID=1000 \
    -e QORUS_GID=100 \
    -v /path/to/omqdir/etc:/opt/qorus/etc \
    -v /path/to/omqdir/log:/opt/qorus/log \
    -v /path/to/omqdir/user:/opt/qorus/user \
    -d dregistry.qoretechnologies.com/qorus:5.1
```

The `--link` option is used to connect the Qorus container to another container. In this case a DB container since Qorus requires a database for functioning.

The `--restart` option configures the behaviour when the container exits, whether it be through an error, a normal exit etc. Since we usually want Qorus to always be running, we select the `always` flag.

The `-p` options are used to forward/publish internal container ports to host's ports.
With the default `options` file configuration, Qorus is set up to run on port 8011 for HTTPS traffic
(note that a self-signed certificate is used).  If you want to have multiple Qorus containers running on
the same machine with the default config, you can simply map them to different outer ports, for example
like this: `-p 12345:8001`. Without these port options, you won't be able to connect to Qorus from outside the container.

The `-e` options set up environment variables. The ones related to system DB configuration are unnecessary if you mount the `/etc` directory containing `options` or `dbparams` file with correctly set up system DB connection.

Likewise, the `-v` options are used for mounting volumes (host filesystem directories) into the container. Mounting them is unnecessary and they can still be accessed easily from the docker data directory.

More info about the database configuration can be seen in the *System DB configuration* section.

## Minimal example for testing

```
docker run --name pg -e POSTGRES_PASSWORD=omq -e TZ=Europe/Prague -e PGTZ=Europe/Prague -d postgres:13
docker run --name qorus --link pg:postgres -p 8011:8011 -e OMQ_DB_NAME=postgres -e OMQ_DB_TYPE=pgsql -e OMQ_DB_HOST=postgres -e OMQ_DB_USER=postgres -e OMQ_DB_PASS=omq -e QORUS_TZ=Europe/Prague -d dregistry.qoretechnologies.com/qorus:5.1
docker exec -it qorus bash
```

## Docker Compose

An alternative way to start Qorus and it's DB is to use Docker Compose. A compose file is included in this repository. To use it, make sure that you have `docker-compose` installed and available in the path. Then simply navigate to this directory and run the following:
```
docker-compose up
```

This will start up the database using `postgres:13` image and also start up a Qorus container that will connect to it.

If you want to launch the containers in background, add `-d` option to the command:
```
docker-compose up -d
```

To check the container state, you can run `docker-compose ps` and to see all the running processes in the containers, you can run `docker-compose top`.

If you want to run additional containers/services or modify any settings, you can simply copy the compose file and edit it as you please.

## Logging

The container's logging is set up to merge stdout and stderr to one log. You can see all the output of the initialization script, the base `qorus` command and of `runit`, which is used as light init system in the Qorus Docker image.

To view log of the container's output, you can use the following:

```
docker logs <CONTAINER>
```

Other Qorus logs are then available in the `OMQ_DIR/log` directory, which you can read either by connecting to the container via bash shell or by mounting the log directory to a directory on the host filesystem.

## Connecting to running container

To connect to running instance with shell access, run the following:

```
docker exec -it <CONTAINER> bash
```

This will open a root bash terminal inside the container.

## Qorus user and directory permissions

The Qorus Docker image is set up so that the `runit` init system runs under *root* user and Qorus runs under a special *qorus* user. By default, this user will have UID and GID both set to 999.

The following directories will also always get "chowned" (get their owner user and group set) to these UID and GID values during the container initialization (before launching Qorus):

* `$OMQ_DIR/etc`
* `$OMQ_DIR/log`
* `$OMQ_DIR/user`

This is done so that the *qorus* user is able to write to these directories.

If you want to have Qorus run under a different user, please create the user in the host system and supply the user's UID and GID to the Docker container with `QORUS_UID` and `QORUS_GID` environment variables. The affected OMQ_DIR directories will then get their owner UID and GID set to these values as well. Be aware that both of these environment variables have to be set, if you want to use them.

Another way how to set up the *qorus* user and group is to mount the `$OMQ_DIR/etc`, `$OMQ_DIR/log` and/or `$OMQ_DIR/user` directories as container volumes, and have them be owned under the correct user and group in the host filesystem. The container will then set up *qorus* user to match this UID and GID, and will also "chown" the other directories if they are owned by someone else.

The order in which these possibilities are checked:

1. If `QORUS_UID` and `QORUS_GID` environment variables are set up, they will be used.
2. If the affected OMQ_DIR directories are owned by other user/group than root (UID or GID different from 0), then UID and GID will be taken from the first directory with different owner than root. These directories are checked in the following order:
    1. OMQ_DIR/etc
    2. OMQ_DIR/log
    3. OMQ_DIR/user
3. The mounted OMQ_DIR directories are either owned by root user and group, or the volumes were not mounted (and are owned by root by default). In this case the default *qorus* user with UID and GID 999 will be used and the directories "chowned" by this user and group.

Make sure to always run the Qorus Docker image under root user (i.e. don't use the `--user` option of `docker run` command), otherwise the above setup will fail.

## Qorus admin account

The *admin* account is created by `schema-tool` during schema alignment during initialization phase of the container. It's password is set to a random string. If `QORUS_ADMIN_PASS` environment variable is set, the account password will be set to it's value.

## Ports

Qorus is set to run on port `8011` for HTTPS traffic by default (using a self-signed certificate). That is, unless
the user changes or mounts their own `options` file in the `$OMQ_DIR/etc` directory. In order to be able to connect
to these ports from outside the container, you have to "publish" these ports using the `-p` option of the
`docker run` command.

## System DB configuration

There are multiple ways to configure Qorus's system DB:

1. Environment variables. Simply set the correct environment variables with DB connection info. For example:

    ```
    -e OMQ_DB_NAME=postgres -e OMQ_DB_TYPE=pgsql -e OMQ_DB_HOST=postgres -e OMQ_DB_USER=postgres -e OMQ_DB_PASS=omq
    ```

    These could also be set from Docker Compose configuration file.

2. Mounting Qorus's `etc` directory containing `options` file with `qorus.systemdb` option set to correct value. For example:

    ```
    qorus.systemdb: pgsql:postgres/omq@postgres(utf8)%postgres
    ```

3. (Deprecated) Mounting `dbparams` file with the `omq` datasource set to correct value. For example:

    ```
    omq=pgsql:postgres/omq@postgres(utf8)%postgres
    ```

    Note that this method is deprecated and might not work in the future releases.

## Environment variables

The following are environment variables that you can set when running the Docker container with the `-e` options, or configure them in Docker Compose file. They are read during container initialization phase, before launching Qorus.

| Variable | Description |
| --- | --- |
| `FORCE_INIT_STEPS` | will force all init steps to be done |
| `FORCE_LOAD_SYSTEM_SERVICES` | will force oloading of all system jobs and services from the `$OMQ_DIR/system` directory |
| `FORCE_UPDATE_ADMIN_USER` | will force the init step of updating admin user password to run; the password won't be updated if the `QORUS_ADMIN_PASS` environment variable isn't set |
| `OMQ_DB_NAME` | database name |
| `OMQ_DB_TYPE` | database type, one of `pgsql`, `oracle` or `mysql` |
| `OMQ_DB_HOST` | database host, assumed to be `localhost` if omitted |
| `OMQ_DB_USER` | database username |
| `OMQ_DB_PASS` | database password |
| `OMQ_DB_ENC` | database encoding, assumed to be `utf8` if omitted |
| `OMQ_DB_TABLESPACE` | DB tablespace, assumed to be `pg_default` if omitted and DB is PostgreSQL type |
| `QORUS_ADMIN_PASS` | Qorus admin user password, to which the *admin* account password will be set |
| `TZ` | The time zone locale for the container (ex: `Europe/Prague`: see https://en.wikipedia.org/wiki/List_of_tz_database_time_zones for more info)
| `QORUS_UID` | UID under which Qorus will run |
| `QORUS_GID` | GID under which Qorus will run |

The environment variables related to system DB configuration are more suited for just trying out the Qorus Docker image. For production use-cases, please consider mounting the `$OMQ_DIR/etc` directory with the `options` file already properly configured.
