#!/bin/sh

__qssh_var_reset() {
    args=
    conn_string=
    remote_env_loc=
    remote_temp_env_loc=
}

__qssh_print_usage() {
    echo "Usage: qssh-${cmd} <SSH-CONNECTION-STRING> [<COMMAND-ARGS> ...]

    If Qorus environment setup file cannot be found, the script will try
    to find location of Qorus OMQ_DIR and create a temporary environment
    file based on the found location. If it cannot find OMQ_DIR location,
    it will simply try to execute the script without any environment
    setup, hoping for automatically set up environment.

    If the environment setup needs to be done, you can help the script find
    OMQ_DIR by setting the \$REMOTE_OMQ_DIR environment variable to location
    of Qorus's OMQ_DIR on the remote server."
}

__qssh_intern_parse_args() {
    if [ "$1" == "-h" -o "$1" == "--help" -o "$1" == "--usage" ]; then
        __qssh_print_usage
        return 1
    fi
    conn_string="$1"
    # remove "conn_string" from positional parameters
    shift 1
    # create argument list and quote each arbgument to ensure it's passed in the same form it's given in
    args=""
    for arg in $@; do
        if [ -z "${args}" ]; then
            args="\"${arg}\""
        else
            args="${args} \"${arg}\""
        fi
    done
}

__qssh_intern_oload_check_simple_option() {
    echo -n "$arg" | grep -E -- "^-[dfhlqvwAEFORSW]+" > /dev/null 2>&1 && return 0
    echo -n "$arg" | grep -E -- "^--(allow-redef|force|full-release|help)" > /dev/null 2>&1 && return 0
    echo -n "$arg" | grep -E -- "^--(no-error-update|override|quiet|reload)" > /dev/null 2>&1 && return 0
    echo -n "$arg" | grep -E -- "^--(show-(defines|full-error|url))" > /dev/null 2>&1 && return 0
    echo -n "$arg" | grep -E -- "^--(use-schedule|verbose|validate)" > /dev/null 2>&1 && return 0
    echo -n "$arg" | grep -E -- "^--(import-cfg-val-rest|sign-schema)" > /dev/null 2>&1 && return 0
    return 1
}

__qssh_intern_oload_check_long_param_option() {
    echo -n "$arg" | grep -E -- "^--(user-)?schema=" > /dev/null 2>&1 && return 0
    echo -n "$arg" | grep -E -- "^--(proxy-)?url=" > /dev/null 2>&1 && return 0
    echo -n "$arg" | grep -E -- "^--(data|index)-ts=" > /dev/null 2>&1 && return 0
    echo -n "$arg" | grep -E -- "^--delete(-id)?=" > /dev/null 2>&1 && return 0
    echo -n "$arg" | grep -E -- "^--(datasource|list|refresh|token)=" > /dev/null 2>&1 && return 0
    echo -n "$arg" | grep -E -- "^--(export-cfg-val|show-release)" > /dev/null 2>&1 && return 0
    return 1
}

__qssh_intern_oload_check_short_param_option() {
    echo -n "$arg" | grep -E -- "^-[prstuDLX]$" > /dev/null 2>&1 && return 0
    return 1
}

__qssh_intern_oload_check_short_param_option_with_arg() {
    echo -n "$arg" | grep -E -- "^-[prstuDLX]" > /dev/null 2>&1 && return 0
    return 1
}

__qssh_intern_oload_parse_args() {
    oload_files=
    oload_opts=
    for arg in $@; do
        # parameter of short option
        if [ "$next_arg_append_opts" = "1" ]; then
            oload_opts="${oload_opts} ${arg}"
            next_arg_append_opts=
            continue
        fi

        # if simple option, add to args and skip
        if __qssh_intern_oload_check_simple_option; then
            oload_opts="${oload_opts} ${arg}"
            continue
        fi

        # if long param option, add to args and skip
        if __qssh_intern_oload_check_long_param_option; then
            oload_opts="${oload_opts} ${arg}"
            continue
        fi

        # if short param option + next arg param, add to oload opts and set helper var for the next arg
        if __qssh_intern_oload_check_short_param_option; then
            oload_opts="${oload_opts} ${arg}"
            next_arg_append_opts=1
            continue
        fi

        if __qssh_intern_oload_check_short_param_option_with_arg; then
            oload_opts="${oload_opts} ${arg}"
            continue
        fi

        # else file -> add to list
        oload_files="${oload_files} ${arg}"
    done
}

__qssh_intern_oload_check_yaml_files() {
    for ofile in ${oload_files}; do
        if echo -n ${ofile} | grep -E "\.y[a]?ml" > /dev/null 2>&1; then
            srcfile=`echo -n $ofile | sed 's/\.yml//' | sed 's/\.yaml//'`
            if [ -e "$srcfile" ]; then
                oload_files="${oload_files} ${srcfile}"
            fi
        fi
    done
    oload_files=`echo $oload_files | tr ' ' '\n' | sort -u | xargs`
}

__qssh_intern_oload_check_resource_files() {
    resource_files=
    for ofile in ${oload_files}; do
        # do not process oload args
        if case ${ofile} in -*) ;; *) false;; esac; then
            continue
        fi
        file_dir=`dirname $ofile`
        echo "Checking resources for file '${ofile}'"

        # parse yaml file
        if echo -n ${ofile} | grep -E "\.y[a]?ml" > /dev/null 2>&1; then
            # skip if the file doesn't contain a resource key
            if ! grep -E "^resource:" ${ofile} > /dev/null 2>&1; then
                continue
            fi

            # parse the file and it's resource section
            resources=
            multiline_item_list=
            multiline_bracket_list=
            while IFS= read -r line || [[ -n "$line" ]]; do
                if [ "$multiline_item_list" = "1" ]; then
                    # check if the line is a list item (begins with dash)
                    if echo -n ${line} | grep -E "^[[:blank:]]*-" > /dev/null 2>&1; then
                        rsrc=`echo -n $line | sed 's/^[[:blank:]]*-[[:blank:]]*//' | sed 's/"//g' | xargs`
                        resources="${resources} ${rsrc}"
                        continue
                    fi
                    break
                fi

                if [ "$multiline_bracket_list" = "1" ]; then
                    # check if there is a closing bracket in the line
                    if echo -n ${line} | grep -E "\]" > /dev/null 2>&1; then
                        # parse until closing bracket
                        rsrc=`echo -n $line | sed 's/[[:blank:]]*\(.*\)\].*/\1/g' | sed 's/,/ /g' | sed 's/"//g' | xargs`
                        resources="${resources} ${rsrc}"
                        break
                    fi

                    # parse line
                    rsrc=`echo -n $line | sed 's/,/ /g' | sed 's/"//g' | xargs`
                    resources="${resources} ${rsrc}"

                elif echo -n ${line} | grep -E "^resource:[[:blank:]]*\[" > /dev/null 2>&1; then
                    # parse the rest of the line after opening bracket
                    rsrc=`echo -n $line | sed 's/^resource:[[:blank:]]*\[//'`

                    # check if there is a closing bracket in the line
                    if echo -n ${rsrc} | grep -E "\]" > /dev/null 2>&1; then
                        # parse until closing bracket
                        rsrc=`echo -n $rsrc | sed 's/\].*//' | sed 's/,/ /g' | sed 's/"//g' | xargs`
                        resources="${resources} ${rsrc}"
                        break
                    fi

                    # we are in the middle of multi-line bracket list
                    multiline_bracket_list=1

                    # parse the resource string from the previously parsed line
                    rsrc=`echo -n $rsrc | sed 's/,/ /g' | sed 's/"//g' | xargs`
                    resources="${resources} ${rsrc}"

                elif echo -n ${line} | grep -E "^resource:[[:blank:]]*$" > /dev/null 2>&1; then
                    # this is a beginning of multiline list (dashes in front of items)
                    multiline_item_list=1
                fi
            done < "${ofile}"

        # parse old-style qsd file
        elif echo -n ${ofile} | grep -E "\.qsd" > /dev/null 2>&1; then
            # skip if the file doesn't contain resource metadata option
            if ! grep -E "#[ ]*resource:" ${ofile} > /dev/null 2>&1; then
                continue
            fi

            # get resource strings from the metadata options
            resources=`sed 's/#[ ]*resource:[ ]*/XXXOLOAD_RSRC:/' $ofile | grep "XXXOLOAD_RSRC:" | xargs | sed 's/XXXOLOAD_RSRC://g'`

        # non-yaml and non-service files
        else
            continue
        fi

        # get full relative paths of the resources
        resources_full=
        for rsrc in ${resources}; do
            resources_full="${resources_full} ${file_dir}/${rsrc}"
        done
        if [ "$resources_full" != "" ]; then
            resources_full=`ls $resources_full 2>/dev/null | xargs`
            echo "Found resources: $resources"
        fi

        # add to resource files list
        resource_files="${resource_files} ${resources_full}"
    done

    # clean up the resource file list
    resource_files=`echo $resource_files | tr ' ' '\n' | sort -u | xargs`
}

__qssh_intern_oload_mkdir() {
    ssh "${conn_string}" "mkdir -p $1"
}

__qssh_intern_oload_mk_tmpdir() {
    __qssh_intern_oload_mkdir "${oload_tmpdir}"
}

__qssh_intern_oload_rm_tmpdir() {
    if ! echo -n "${oload_tmpdir}" | grep "^/tmp/" > /dev/null 2>&1; then
        echo "Error: Aborting before removing non-temp dir: '${oload_tmpdir}'" >&2
        return 1
    fi
    ssh "${conn_string}" "rm -rf ${oload_tmpdir}"
}

__qssh_intern_oload_cleanup() {
    __qssh_intern_oload_rm_tmpdir
    if [ "$remote_temp_env_loc" != "" ]; then
        ssh "${conn_string}" "rm -f ${remote_temp_env_loc}"
    fi
}

__qssh_intern_connection_works() {
    ssh "${conn_string}" "echo -n"
}

__qssh_intern_check_connection() {
    if [ "$conn_string" == "" ]; then
        echo "Error: Connection to server not set." >&2
        __qssh_print_usage
        return 1
    fi
    if ! __qssh_intern_connection_works; then
        echo "Error: Cannot connect to '${conn_string}'" >&2
        return 1
    fi
}

__qssh_intern_check_remote_env_config() {
    echo "Checking location of environment setup file."
    env_test_locations="/opt/qorus/bin/env.sh /opt/qorus-initweb/env.sh /etc/profile.d/qorus.sh"
    for env_loc in ${env_test_locations}; do
        if ssh "${conn_string}" "test -e ${env_loc}"; then
            remote_env_loc="${env_loc}"
            break
        fi
    done

    if [ "$remote_env_loc" == "" ]; then
        echo "Warning: Could not find location of environment setup file on the remote server." >&2
    else
        echo "Environment setup file found in '${remote_env_loc}'"
        return 0
    fi

    # try to find location of Qorus
    echo "Trying to find location of Qorus OMQ_DIR."
    qorus_test_locations="/opt/qorus /srv/qorus /home/qorus/omq /home/qorus/omqdir ${OMQ_DIR} ${REMOTE_OMQ_DIR}"
    for qorus_loc in ${qorus_test_locations}; do
        if ssh "${conn_string}" "test -d ${env_loc}"; then
            remote_qorus_loc="${qorus_loc}"
            break
        fi
    done

    if [ "$remote_qorus_loc" == "" ]; then
        echo "Warning: Could not find location of Qorus OMQ_DIR on the remote server." >&2
        return 0
    else
        echo "Qorus OMQ_DIR found in '${remote_qorus_loc}'"

        # create temp environment file on remote server
        echo "Creating temporary environment setup file on the remote server."
        remote_temp_env_loc="/tmp/.qssh-qorus-env.sh"

        __env_file_content="export OMQ_DIR=${remote_qorus_loc}
        export PATH=${remote_qorus_loc}/bin:${remote_qorus_loc}/../bin:\\\$PATH
        export LD_LIBRARY_PATH=${remote_qorus_loc}/lib:${remote_qorus_loc}/../lib:${remote_qorus_loc}/../lib/x86_64-linux-gnu:${remote_qorus_loc}/../lib64
        export QORE_INCLUDE_DIR=${remote_qorus_loc}/qlib
        export QORE_MODULE_DIR=${remote_qorus_loc}/qlib
        export QORE_DATA_PROVIDERS=QorusDataProviders
        export QORE_CONNECTION_PROVIDERS=QorusConnectionProvider
        export QORE_DATASOURCE_PROVIDERS=QorusDatasourceProvider"

        ssh "${conn_string}" "echo \"${__env_file_content}\" > ${remote_temp_env_loc}; chmod +x ${remote_temp_env_loc}"
        remote_env_loc="${remote_temp_env_loc}"
    fi
}

__qssh_intern_check_args() {
    __qssh_intern_check_connection && \
    __qssh_intern_check_remote_env_config
}

__qssh_intern_print_args() {
    echo "Will execute ${cmd} via connection: '${conn_string}'"
    echo "Arguments: '${args}'" && echo "-----"
}

__qssh_intern_exec_ssh_cmd() {
    if [ "$remote_env_loc" == "" ]; then
        ssh "${conn_string}" "${cmd} ${args}"
    else
        ssh "${conn_string}" ". ${remote_env_loc} && ${cmd} ${args}"
    fi
}

__qssh_intern_do_cmd() {
    __qssh_var_reset
    __qssh_intern_parse_args $@ && \
    __qssh_intern_check_args && \
    __qssh_intern_print_args && \
    __qssh_intern_exec_ssh_cmd
}

qssh-ocmd() {
    cmd=ocmd
    __qssh_intern_do_cmd $@
}

qssh-ojview() {
    cmd=ojview
    __qssh_intern_do_cmd $@
}

qssh-oload() {
    cmd=oload
    __qssh_var_reset
    __qssh_intern_parse_args $@ || return 1
    __qssh_intern_check_args || return 1
    __qssh_intern_print_args || return 1

    # parse args for options and loaded files
    oload_files=
    __qssh_intern_oload_parse_args ${args} || return 1

    # check for yaml files and add their source files if they are missing
    __qssh_intern_oload_check_yaml_files

    # check service files for resources
    __qssh_intern_oload_check_resource_files

    # create temp dir on remote server for copying oloaded files
    oload_tmpdir=/tmp/.qssh-oload-`head -c 192 /dev/urandom | tr -dc A-Za-z0-9 | head -c 24`
    if ! __qssh_intern_oload_mk_tmpdir; then
        echo "Error: Could not create temporary directory on remote server '${conn_string}'." >&2
        return 1
    fi

    # copy resource files to temp dir on remote server
    for rsrcfile in ${resource_files}; do
        echo "Copying resource file '${rsrcfile}' to remote server"

        # create matching directory structure for the file inside the temp dir
        file_dir=`dirname $rsrcfile`
        if [ "$file_dir" == "." ]; then
            file_dir=
        else
            if ! echo -n "$file_dir" | grep -E "^/" > /dev/null 2>&1; then
                file_dir="/${file_dir}"
            fi
            __qssh_intern_oload_mkdir "${oload_tmpdir}${file_dir}"
        fi

        # copy the file into the tempdir
        if ! scp "${rsrcfile}" "${conn_string}:${oload_tmpdir}${file_dir}/"; then
            echo "Error: Could not copy resource file '${rsrcfile}' to remote server." >&2
            __qssh_intern_oload_cleanup
            return 1
        fi
    done

    # copy loaded files to temp dir in container
    oload_files_final=
    for ofile in ${oload_files}; do
        # do not copy oload args
        if case ${ofile} in -*) ;; *) false;; esac; then
            # add arg
            oload_files_final="${oload_files_final} ${ofile}"
            continue
        fi
        echo "Copying file '${ofile}' to remote server"

        # create matching directory structure for the file inside the temp dir
        file_dir=`dirname $ofile`
        if [ "$file_dir" == "." ]; then
            file_dir=
        else
            if ! echo -n "$file_dir" | grep -E "^/" > /dev/null 2>&1; then
                file_dir=/${file_dir}
            fi
            __qssh_intern_oload_mkdir "${oload_tmpdir}${file_dir}"
        fi

        # copy the file into the tempdir
        if ! scp "${ofile}" "${conn_string}:${oload_tmpdir}${file_dir}/"; then
            echo "Error: Could not copy file '${ofile}' to remote server." >&2
            __qssh_intern_oload_cleanup
            return 1
        fi
        oload_files_final="${oload_files_final} ${oload_tmpdir}${file_dir}/`basename $ofile`"
    done
    oload_files_final=`echo -n "$oload_files_final" | xargs`

    echo "Executing ${cmd}..."
    echo "-----"

    # execute oload
    if [ "$remote_env_loc" == "" ]; then
        ssh "${conn_string}" "cd ${oload_tmpdir} && ${cmd} ${oload_opts} ${oload_files_final}"
        ret=$?
    else
        ssh "${conn_string}" ". ${remote_env_loc} && cd ${oload_tmpdir} && ${cmd} ${oload_opts} ${oload_files_final}"
        ret=$?
    fi

    # remove temp dir with files and temp env file
    __qssh_intern_oload_cleanup
    return ${ret}
}

qssh-oprop() {
    cmd=oprop
    __qssh_intern_do_cmd $@
}

qssh-ostart() {
    cmd=ostart
    __qssh_intern_do_cmd $@
}

qssh-ostatus() {
    cmd=ostatus
    __qssh_intern_do_cmd $@
}

qssh-ostop() {
    cmd=ostop
    __qssh_intern_do_cmd $@
}

qssh-oview() {
    cmd=oview
    __qssh_intern_do_cmd $@
}

qssh-qctl() {
    cmd=qctl
    __qssh_intern_do_cmd $@
}

qssh-qdbg() {
    cmd=qdbg
    __qssh_intern_do_cmd $@
}

qssh-qdbg-remote() {
    cmd=qdbg-remote
    __qssh_intern_do_cmd $@
}

qssh-qrest() {
    cmd=qrest
    __qssh_intern_do_cmd $@
}

qssh-qbugreport() {
    cmd=qbugreport
    __qssh_intern_do_cmd $@
}

qssh-schema-tool() {
    cmd=schema-tool
    __qssh_intern_do_cmd $@
}

qssh-user-tool() {
    cmd=user-tool
    __qssh_intern_do_cmd $@
}

# qore commands
qssh-rest() {
    cmd=rest
    __qssh_intern_do_cmd $@
}

qssh-schema-reverse() {
    cmd=schema-reverse
    __qssh_intern_do_cmd $@
}

qssh-sfrest() {
    cmd=sfrest
    __qssh_intern_do_cmd $@
}

qssh-soaputil() {
    cmd=soaputil
    __qssh_intern_do_cmd $@
}

qssh-sqlutil() {
    cmd=sqlutil
    __qssh_intern_do_cmd $@
}

qssh-qdp() {
    cmd=qdp
    __qssh_intern_do_cmd $@
}

qssh-saprest() {
    cmd=saprest
    __qssh_intern_do_cmd $@
}
