#!/usr/bin/env python3

import argparse
import datetime
import getpass
import glob
import os
import re
import shutil
import socket
import sys
import tarfile
import tempfile

LoadFileTypes = {
    "qfd": True,
    "qsd": True,
    "java": True,
    "qclass": True,
    "qconst": True,
    "qwf": True,
    "qjob": True,
    "qconn": True,
    "qsm": True,
    "qmapper": True,
    "qvmap": True,
    "qscript": True,
    "qstep": True,
    "qmc": True,
    "yaml": True,
    "py": True
}

# file extensions to be ignored when packaging a release
ExtraFileTypes = {
    "wsdl": True,
    "xml": True,
    "xsd": True,
    "dtd": True,
    "qm": True,
    "qlib": True,
    "jar": True,
    "class": True,
    "json": True,
    "qtest": True
}


class MakeRelease:
    @staticmethod
    def getLabel(prefix, label):
        # check for file name instead of label
        if os.path.isfile(label):
            print >> sys.stderr, "ERROR: label '{}' is an existing file - check arguments and try again".format(label)
            sys.exit(1)

        if label[:len(prefix)] == prefix:
            return label

        return prefix + label

    @staticmethod
    def fixPrefix(string):
        # substitute repeated '/' with a single '/'
        string = re.sub('//+', '/', string)
        # remove leading '/'
        string = string[1:] if string[0] == '/' else string
        # remove trailing '/'
        string = string[:-1] if string[-1] == '/' else string

        return string

    @staticmethod
    def getPythonDestDir():
        return "user/python/lib/python{}.{}/site-packages".format(sys.version.major, sys.version.minor)

    @staticmethod
    def mkdir(d, msg='', mode=0o777):
        try:
            os.mkdir(d, mode)
            print("created: {} '{}'".format(msg, d))
        except e:
            error("mkdir({}) failed: {}".format(d, e))

    @staticmethod
    def getUniqueDirname():
        return "make-release-{}-{}-pid-{}".format(socket.gethostname(), getpass.getuser(), os.getpid())

    @staticmethod
    def doCmd(fmt):
        cmd = '{} {}'.format(fmt, argv)
        if self._opts.verbose:
            print("exec: {}".format(cmd))
        rc = os.system(cmd)
        if rc > 0:
            raise Exception("SYSTEM-ERROR", "command: {} returned error code {}".format(cmd, rc))

    @staticmethod
    def error(fmt):
        print("ERROR: {} {}".format(fmt, sys.argv[1:]))
        sys.exit(1)

    @staticmethod
    def is_readable(path):
        try:
            with open(path) as f:
                return f.readable()
        except IOError as err:
            return False

    def __init__(self):
        self._slist = []
        self._ilist = []
        self._ulist = []
        self._rname = ''
        self._tar = ''
        
        # set release dir
        self._rdir = os.getenv('QORUS_RELEASE_DIR', os.path.join(os.getenv('HOME'), "releases"))
        
        # set options
        _parser = argparse.ArgumentParser()
        _parser.add_argument('-u', '--user', help='add user code to system release', dest='user')
        _parser.add_argument('-U', '--user-src', dest='usrc')
        _parser.add_argument('-m', '--mod', dest='fix-module-paths')
        _parser.add_argument('-p', '--prefix', dest='pref')
        _parser.add_argument('-P', '--add-prefix', dest='padd')
        _parser.add_argument('-r', '--release-dir', dest='rdir')
        _parser.add_argument('-q', '--user-sql', dest='usql')
        _parser.add_argument('-R', '--show-release-dir', dest='showr', action='store_true')
        _parser.add_argument('-c', '--compress', dest='comp', action='store_true')
        _parser.add_argument('-f', '--refresh', dest='ref', action='store_true')
        _parser.add_argument('-C', '--refresh-compat', dest='rcompat', action='store_true')
        _parser.add_argument('-F', '--full-release', dest='full', action='store_true')
        _parser.add_argument('-i', '--install', dest='inst', action='store_true')
        _parser.add_argument('--keep', dest='keep', action='store_true')
        _parser.add_argument('-v', '--verbose', dest='verbose')
        _parser.add_argument('--qorusdir', dest='qorusdir')
        _parser.add_argument('-a', '--python-modules', dest='pymoddir', action='store_true')
        _parser.add_argument('-b', '--python-dest', dest='pymoddest', action='store_true')
        _parser.add_argument('label')

        self._opts, self._args = _parser.parse_known_args()
        print(self._opts.label)
        print(self._opts)
        print(self._args)

        if self._opts.showr:
            print("{}".format(self._rdir))
            sys.exit(1)

        # set user label
        if self._opts.user:
            self._opts.user = MakeRelease.getLabel("qorus-user-", self._opts.user)

        if not self._opts.usrc and (self._opts.pref or self._opts.padd):
            self._opts.usrc = "."

        # fix prefix
        if self._opts.pref:
            self._opts.pref = fixPrefix(self._opts.pref)

        if self._opts.padd:
            self._opts.padd = fixPrefix(self._opts.padd)

        # if no user source listed, then set to current directory
        if not self._opts.usrc:
            self._opts.usrc = "."

        # set pymoddir/pymoddest
        if self._opts.pymoddir:
            if not self._opts.pymoddest:
                self._opts.pymoddest = getPythonDestDir()
        elif self._opts.pymoddest:
            print("python-module-dest can only be used in addition to the python-module option")



    def checkAbsolutePath(self, path):
        if os.path.isabs(path):
            print >> sys.stderr, "ERROR: release file component {} is an absolute path; must be a relative path from {}".format(path, self._opts.usrc)
            sys.exit(1)


    def doFile2(self, file_list, d, path):
        self.checkAbsolutePath(path)

        # skip backup files
        if re.search('~$', path):
            return

        # skip release files
        if re.search('\.qrf$', path):
            print('skipping release file \'{}\''.format(path))
            return

        print('d', d)
        print('path', path)
        if os.path.isdir(os.path.join(d, path)):
            # skip files in "old" directories
            if re.search('old$', path):
                return
            for path1 in glob.glob(os.path.join(os.path.join(d, path), "*")):
                self.doFile2(file_list, d, os.path.join(path, os.path.basename(path1)))
            return

        file_list += path


    def doFile(self, path):
        self.checkAbsolutePath(path)

        # skip backup files
        if re.search('~$', path):
            return

        d = self._opts.usrc
        self.doFile2(self._ulist, self._opts.usrc, path)

        new_path = os.path.join(d, path)
        if not MakeRelease.is_readable(new_path):
            print("ERROR: cannot find file '{}'".format(new_path))
            sys.exit(1)


    def checkFiles(self, args):
        # make file lists
        self._slist = self._ilist = self._ulist = []

        # and check to see if all files exist
        for path in args:
            if re.search('[\*\?]', path):
                paths = glob.glob(os.path.join(self._opts.usrc, path))
                if not paths:
                    print >> sys.stderr, "ERROR: path {} does not match any files".format(path)
                    sys.exit(1)

                map(self.doFile, paths)
                continue

            self.doFile(path)


    def makeList(self, file_list, resource_list):
        # make a set of the file list for quick lookups
        file_map = { i : True for i in file_list }

        string = ""
        for entry in file_list:
            if re.search('~$', entry):
                continue

            if re.search('[\*\?]', entry):
                for e in glob.glob(entry):
                    self.processFile(e, file_map, string, resource_list)
            else:
                self.processFile(entry, file_map, string, resource_list)

        # remove trailing space
        string.strip()
        return string


    def processFile(self, entry, file_map, string, resource_list):
        string += '{} '.format(entry)

        if re.search('\.yaml$', entry):
            try:
                # try to parse the YAML and see if it has a code reference that can be added to the release
                with open(entry) as e:
                    fh = yaml.full_load(e)
                    if 'code' in fh:
                        src = fh['code']
                        src = os.path.join(os.path.dirname(entry), src)
                        if MakeRelease.is_readable(src) and src not in file_map:
                            file_map[src] = True
                            string += '{} '.format(src)
            except:
                pass

        # add service resources to releasecd 
        if resource_list and re.search('\.qsd$', entry):
            rlist = []
            with open(entry) as e:
                lines = e.readLines()
            
            for line in lines:
                resource_name = re.match('#[[:blank:]]*(resource|template|bin-resource|text-resource)[[:blank:]]*:[[:blank:]]*(.+)/)[1]', line)
                
                if resource_name:
                    resource_name = resource_name.strip()
                    self.checkAbsolutePath(resource_name)
                    d = os.path.dirname(entry)
                    if d != ".":
                        resource_name = os.path.join(d, resource_name)
                    if not re.search('[\*\?]', resource_name) and not MakeRelease.is_readable(resource_name):
                        error("service {} references resource {} that does not exist ({})".format(entry, resource_name, strerror()))
                    rlist += resource_name

            for r in rlist:
                resource_list.append(('source', r))


    def delTree(self, targ):
        shutil.rmtree(targ)


    def doCreateTar(self, opt, tarf, targ):
        with tarfile.open(tarf, 'w:{}'.format(opt)) as tar:
            for ta in targ:
                tar.add(ta)


    def doExtractTar(self, opt, tarf, topt):
        with tarfile.open(tarf, 'r:{}'.format(opt)) as tar:
            tar.extractAll(path=topt)


    def copyFile(self, source, target, quote_source=True):
        shutil.copyfile(source, target)

    def logDebug(self, fmt):
        if self._opts.verbose > 1:
            print('{} {}'.format(fmt), argv)

    def doResources(self, resource_list, dir_name):
        dh = {}
        for rh in resource_list:
            d = os.path.dirname(rh['source'])
            fn = os.path.basename(rh['source'])

            tdir = normalize_dir(dir_name if d == "." else os.path.join(dir_name, d))
            if (tdir not in dh):
                if not os.path.isdir(tdir):
                    MakeRelease.mkdir(tdir)

                dh[tdir] = True

            # include all files in the subdirectory matching the pattern
            if re.search('[\*\?]', fn):
                dc = os.path.dirname(fn)
                if (dc != "."):
                    fn = os.path.basename(fn)
                    d += os.path.sep + dc

                doGlob(d, fn, tdir)
            else:
                self.copyFile(rh.source, tdir + os.path.sep)


    def doGlob(self, d, fn, tdir):
        for fstr in glob(os.path.join(d, fn)):
            # skip files ending in "~" as backup files
            if re.search('~$', fstr):
                continue

            if os.path.isdir(fstr):
                doGlob(d, os.path.join(fstr.substr(d.size() + 1), basename(fn)), tdir)
                continue

            # skip special files, sockets, device files, etc
            if not os.path.isfile(fstr):
                continue

            targ = tdir + os.path.sep
            dir_name = os.path.dirname(fn)
            if (dir_name != "."):
                targ += dir_name + os.path.sep
                if not os.path.isdir(targ):
                    MakeRelease.mkdir(targ)

            self.copyFile(fstr, targ)


    def createUserReleaseFile(self, path):
        # create release file
        with open(path, 'w') as f:
            f.write("# automatically generated by {} on {} ({}@{})\n".format(os.path.basename(__file__), datetime.datetime.now(), os.getenv('USER'), socket.gethostname()))

            root_dir = os.path.dirname(path)
            if root_dir == ".":
                root_dir = ""

            for fn in self._ulist:
                # check for known file extensions
                ext = MakeRelease.getExt(fn)
                if not ext:
                    # see if it's an executable
                    if not os.access(fn, os.X_OK):
                        print("warning: no extension in file '{}'...".format(fn))
                elif ext in LoadFileTypes:
                    f.write("load {}\n".format(getLoadPath(fn, root_dir)))
                elif ext not in ExtraFileTypes:
                    print("warning: unknown extension '{}' in file '{}'...".format(ext, fn))

            # now add user sql files
            for fn in (self._opts.usql):
                # check for known file extensions
                ext = MakeRelease.getExt(fn)

                if not re.search('sql$', ext):
                    print("warning: user SQL file extension is not 'sql': {}".format(fn))

                f.write("omquser-exec-sql {}\n".format(getLoadPath(fn, root_dir)))

            if self._opts.ref:
                f.write("refresh-recursive\n")
            elif self._opts.rcompat:
                f.write("refresh-all\n")

        os.chmod(path, 0o644)
        print("created user release file %s\n".format(path))


    def getLoadPath(self, fn, root_dir):
        if fn.startswith(root_dir):
            fn = fn[root_dir.size():]
        if self._opts.pref:
            return os.path.join(self._opts.pref, fn)
        if self._opts.padd:
            return os.path.join(self._opts.padd, fn)
        return fn


    @staticmethod
    def getExt(fn):
        return re.match('.*\.(.*)$/)[0]', fn)


    def exec(self):
        if not self._args:
            print("ERROR: no files given on the command line")
            usage()

        self.checkFiles(self._args)

        label = MakeRelease.getLabel("qorus-user-", self._opts.label)

        # check release directory
        if not os.path.isdir(self._rdir):
            MakeRelease.mkdir(_rdir, "root release directory")

        rname = os.path.join(self._rdir, label)
        rname = os.path.normpath(rname)

        ulabel = self._opts.user

        if not os.path.isdir(rname):
            MakeRelease.mkdir(rname, "release directory")

        if not os.path.isfile(os.path.join(rname, "install.sh")):
            pathname = os.path.join(os.path.join(os.getenv('OMQ_DIR'), "templates"), "install.sh")

            # copy install.sh to new directory
            self.copyFile(pathname, os.path.join(rname, "install.sh"))
            print("copied:{} to:{}".format(pathname, os.path.join(rname, "install.sh")))

        if not os.path.isdir(os.path.join(rname, "releases")):
            # create releases subdirectory
            if MakeRelease.mkdir(os.path.join(rname, "releases")):
                error("MakeRelease.mkdir({}) failed: {}".format(os.path.join(rname, "releases"), strerror()))

        if self._ulist:
            tar_file_name = "%s.tar.gz".format(ulabel)
            user_file_list = self._ulist
            if self._opts.usql:
                user_file_list += self._opts.usql

            # path -> info
            resource_list = []
            
            file_list = self.makeList(user_file_list, resource_list)

            if not self._opts.pref and not self._opts.padd and resource_list:
                self._opts.padd = "user"

            # create release in prefix directory
            if self._opts.pref:
                pdir = self._opts.pref

                # get list of paths
                if os.name == 'Windows':
                    pdir = regex_subst(pdir, "/", "\\", RE_Global)

                path_component_list = split(os.path.sep, pdir.split)

                # get unique directory prefix name
                unique_dir = getUniqueDirname()
                dir_name = os.path.join(os.path.join(tempfile.gettempdir(), unique_dir), pdir)

                # create install directory
                MakeRelease.mkdir(dir_name)

                # on_exit:
                #     if not self._opts.keep:
                #         os.rmdir(os.path.join(tempfile.gettempdir(), unique_dir))


                logDebug("changing to CWD: %y", self._opts.usrc)
                #chdir(self._opts.usrc)
                # copy files to temporary release directory
                self.copyFile(file_list, dir_name, False)
                doResources(resource_list, dir_name)

                # create tar file
                #chdir(tempfile.gettempdir()() + DirSep + unique_dir)
                self.doCreateTar("gz", os.path.join(rname, tar_file_name), path_component_list[0])
            elif self._opts.padd:
                pdir = self._opts.padd

                # get unique directory prefix name
                unique_dir = getUniqueDirname()

#%ifdef Windows
                pdir = replace(pdir, "/", "\\")
#%endif
                # get list of path components
                path_component_list = split(os.path.sep, pdir)

                # create install directory
                dir_name = os.path.join(os.path.join(tempfile.gettempdir(), unique_dir), pdir)
                MakeRelease.mkdir(dir_name)

                logDebug("creating install dir: %y", dir_name)

                # on_exit:
                #     if (!self._opts.keep)
                #         delTree(tempfile.gettempdir()() + DirSep + unique_dir)


                logDebug("changing to CWD: %y", self._opts.usrc)

                #chdir(self._opts.usrc)
                doResources(resource_list, dir_name)

                # copy files to temporary release directory
                # create temporary tar file
                tmp_tar = os.path.join(tempfile.gettempdir(), os.path.join(unique_dir, "tqr.tar.gz"))
                self.doCreateTar("gz", tmp_tar, file_list)

                # unpack release in new position
                #chdir(dir_name)
                doExtractTar("gz", tmp_tar)
                unlink(tmp_tar)

                # move module files to new location
                if (self._opts.mod):
                    target_dir
                    for fn in self._ulist:
                        if re.search('\.qm$', fn):
                            if not target_dir:
                                target_dir = os.path.join(tempfile.gettempdir(), os.path.join(unique_dir, "user/modules"))
                                MakeRelease.mkdir(target_dir, mode=0o755)

                            # move module file to target
                            targ = os.path.join(target_dir, basename(fn))
                            rename(fn, targ)

                # create tar file
                #chdir(tempfile.gettempdir()() + DirSep + unique_dir)
                self.doCreateTar("gz", os.path.join(rname, tar_file_name), path_component_list[0])
            else:
                #chdir(self._opts.usrc)
                self.doCreateTar("gz", os.path.join(rname, tar_file_name), file_list)

            print("created user tar file {}/{}\n".format(rname, tar_file_name))

            if (self._opts.pymoddir):
                #chdir(rname)
                temp = getUniqueDirname()
                doCmd("MakeRelease.mkdir -p %s", temp)
                doExtractTar("gz", tar_file_name + " -C " + temp)
                dest = os.path.join(rname, os.path.join(temp, self._opts.pymoddest))
                doCmd("MakeRelease.mkdir -p %s", dest)
                doCmd("cp -r %s %s", self._opts.pymoddir, dest + DirSep + basename(self._opts.pymoddir))
                #chdir(temp)
                self.doCreateTar("gz", os.path.join(rname, tar_file_name), "*")
                delTree(os.path.join(rname, temp))
                print("Adding python module at destination: {}\n".format(self._opts.pymoddest))

            # create release file
            release_file_namef = "{}/releases/{}.qrf".format(rname, ulabel)
            self.createUserReleaseFile(release_file_namef)

            if self._opts.comp:
                # save current working directory
                #logDebug("rname: %y ut: %y\n", rname, tar_file_name)
                #chdir(dirname(rname))
                # make compressed release tar.bz2; exclude backup files
                tar = label + ".tar.bz2"
                self.doCreateTar("bz2", tar, "--exclude=\"*~\" " + label)
                print("create release archive: {}/{}\n".format(getcwd(), tar))


            if (self._opts.inst):
                #chdir(rname)

                self._opts

                if self._opts.verbose:
                    self._opts += " -v"
                if self._opts.full:
                    self._opts += " -F"

                doCmd("sh ./install.sh" + self._opts)

        print("done!")


    def usage():
        py_dest_dir = getPythonDestDir()

        print("""usage: {} [options] <label> <files>
usage: {} <options> <files>
   -h,--help                show this help text
   -a,--python-modules=ARG  package the given directory as a Python module dir
                            to be installed in the target under
                            $OMQ_DIR/{}
                            (use -b to override the target dir)
   -b,--python-dest=ARG     store files packaged with -a in the given directory
                            under $OMQ_DIR/
   -c,--compress            make a compressed tar file of the release
   -C,--refresh-compat      include old commands to refresh objects after
                            loading
   -f,--refresh             include commands to refresh objects after loading
   -F,--full-release        verify release completeness, only with -i
   -i,--install             exec install.sh after packaging (user rel. only)
      --keep                do not delete temporary packaging directory
   -p,--prefix=ARG          sets user prefix directory for relative paths in the
                            target filesystem (implies -U.; note: makes a flat
                            release in this dir)
   -P,--add-prefix=ARG      prepends a prefix dir for relative paths in the
                            target filesystem (implies -U.)
   -m,--fix-module-paths    install Qore modules in user/modules
   -q,--user-sql=ARG        adds an SQL file to execute in the omquser schema
   -r,--release-dir=ARG     set release directory (def: {})
   -R,--show-release-dir    show the release directory and exit (def: {})
   -u,--user=ARG            
   -U,--user-src=ARG        sets the root source directory for release
                            components; all release componets should be given as
                            relative paths from this directory
   -v,--verbose             output more information
      --qorusdir            set the Qorus src directory

Examples:
  {}  -U. mylabel services/*.qsd
       It creates a user-code release with all service files
       in the default release-dir/mylabel directory.
  {}  -U. -lmylabel services/*.qsd
       It creates only the load script manifest for service files
       named mylabel.qrf
\n""".format(os.path.basename(__file__), os.path.basename(__file__), py_dest_dir,
             self._rdir, self._rdir, os.path.basename(__file__), os.path.basename(__file__)))
        
        sys.exit(1)


globals = {}
if __name__ == "__main__":
    makeRelease = MakeRelease()
    makeRelease.exec()